#include "card.h"
#include "requests.h"
#include "state.h"

#include <ctype.h>
#include <curses.h>
#include <errno.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

static int is_input_pending (void);
static void init_signals (void);
static void init_curses (void);
static void determine_window_size (void);
static void resize (int sig);

static int is_fkey (int ch);
static int my_unctrl (int ch);

static int process_pending_requests (struct state *cur_state);
static void handle_input (struct state *cur_state, int ch);

int main (int argc, char *argv[])
{
    struct state *s;

    s = state_new ();
    state_read_cardfile (s, argv[1]);

    init_signals ();
    init_curses ();
    determine_window_size ();

    post_request (REQUEST_FORCE_REFRESH);

    while (process_pending_requests (s))
    {
        int ch;

        do
        {
            errno = 0;
            ch = getch ();
            if (ch == ERR && errno == EINTR)
                /* signal interrupted */
                continue;

            handle_input (s, ch);
        }
        while (is_input_pending ());
    }

    endwin ();
    state_free (s);
    return 0;
}

static int is_input_pending (void)
{
    int ch;

    nodelay (stdscr, 1);

    do
    {
       errno = 0;
       ch = getch ();
    }
    while (ch == ERR && errno == EINTR);

    nodelay (stdscr, 0);

    if (ch == ERR)
        /* No input pending */
        return 0;

    ungetch (ch);
    return 1;
}

static void init_signals (void)
{
    signal (SIGWINCH, resize);
    siginterrupt (SIGWINCH, 1);
}

static void init_curses (void)
{
    initscr ();
    /* This version doesn't use color - but if color is added back in,
       a call to if has_colors () start_color, plus init_pair goes here. */
    keypad (stdscr, TRUE);
    raw ();
    noecho ();
    typeahead (-1);
}

static void determine_window_size (void)
{
    /* Standard curses trick for getting it to recognize the resize */
    endwin ();
    refresh ();

    post_request_i (REQUEST_INFORM_HEIGHT, LINES);
    post_request_i (REQUEST_INFORM_WIDTH, COLS);
}

static void resize (int sig)
{
    post_signal_request (REQUEST_RESIZE);
}

static int is_fkey (int ch)
{
    return ch > 255;
}

static int my_unctrl (int ch)
{
    return unctrl ((chtype) ch) [1];
}

/* Returns false if quit requested */
static int process_pending_requests (struct state *cur_state)
{
    /* Exit loop if input pending. The input may request another screen
       refresh, in which case some of our work queue should be skipped */
    while (any_requests () && ! is_input_pending ())
    {
        int msg, i;
        void *vp;

        pop_request (&msg, &i, &vp);

        switch (msg)
        {
        case REQUEST_RESIZE:
            determine_window_size ();
            post_request (REQUEST_FORCE_REFRESH);
            break;
        case REQUEST_DUPLICATE_CARD:
            state_duplicate_card (cur_state);
            break;
        case REQUEST_INSERT:
            state_insert (cur_state, (char) i);
            break;
        case REQUEST_PREVIOUS_CARD:
            /* If the screen will be repainted, don't bother right now */
            if (pending_requests_of_type (REQUEST_PREVIOUS_CARD) > 1)
                state_previous_card (cur_state, NO_REFRESH);
            else
                state_previous_card (cur_state, REFRESH_OK);
            break;
        case REQUEST_NEXT_CARD:
            /* If the screen will be repainted, don't bother right now */
            if (pending_requests_of_type (REQUEST_NEXT_CARD) > 1)
                state_next_card (cur_state, NO_REFRESH);
            else
                state_next_card (cur_state, REFRESH_OK);
            break;
        case REQUEST_MOVE_CURSOR:
            state_move_cursor (cur_state, i);
            break;
        case REQUEST_NEW_CARD:
            state_new_card (cur_state);
            break;
        case REQUEST_FORCE_REFRESH:
            state_force_refresh (cur_state);
            break;
        case REQUEST_SAVE:
            state_save (cur_state);
            break;
        case REQUEST_DELETE_PREV:
            state_delete_prev (cur_state);
            break;
        case REQUEST_DELETE_AT:
            state_delete_at (cur_state);
            break;
        case REQUEST_INFORM_HEIGHT:
            state_inform_height (cur_state, i);
            break;
        case REQUEST_INFORM_WIDTH:
            state_inform_width (cur_state, i);
            break;
        case REQUEST_DELETE_CARD:
            state_delete_card (cur_state);
            break;
        case REQUEST_DIALOG_YES:
            state_dialog_yes (cur_state);
            break;
        case REQUEST_DIALOG_NO:
            state_dialog_no (cur_state);
            break;
        case REQUEST_DIALOG_ACCEPT:
            state_dialog_accept (cur_state);
            break;
        case REQUEST_DIALOG_ABORT:
            state_dialog_abort (cur_state);
            break;
        case REQUEST_GRACEFUL_EXIT:
            state_graceful_exit (cur_state);
            break;
        case REQUEST_PDF:
            state_pdf (cur_state);
            break;
        case REQUEST_EXIT_NOW:
            return FALSE;
        }
    }

    return TRUE;
}

static void handle_input (struct state *cur_state, int ch)
{
    /* We've asked the user yes or no. Only respond to yes, no and escape/ctrl-c */
    if (cur_state->input_mode == DIALOG_YN)
    {
        if (iscntrl (ch) && ! is_fkey (ch))
        {
            switch (my_unctrl (ch))
            {

            case 'C':
            case 'X':
            case '[':   post_request (REQUEST_DIALOG_ABORT);
                        return;

            default:    return;
            }
        }

        switch (ch)
        {
        case 'Y':
        case 'y':   post_request (REQUEST_DIALOG_YES);
                return;

        case 'N':
        case 'n':   post_request (REQUEST_DIALOG_NO);
                    return;

        default:    return;
        }
    } /* DIALOG_YN */

    /* We've asked the user for a string input at the bottom of screen */
    if (cur_state->input_mode == DIALOG_STRING)
    {
        if (iscntrl (ch) && ! is_fkey (ch))
        {
            switch (my_unctrl (ch))
            {

            case 'C':
            case 'X':
            case '[':   post_request (REQUEST_DIALOG_ABORT);
                        return;

            /* Due to legacy terminal issues with KEY_BACKSPACE */
            case '?':   post_request (REQUEST_DELETE_PREV);
                        return;

            case 'J':   post_request (REQUEST_DIALOG_ACCEPT);
                        return;

            default:    return;
            }
        }

        switch (ch)
        {
        case KEY_LEFT:    post_request_i (REQUEST_MOVE_CURSOR, LEFT);
                return;

        case KEY_RIGHT:    post_request_i (REQUEST_MOVE_CURSOR, RIGHT);
                return;

        case KEY_BACKSPACE:    post_request (REQUEST_DELETE_PREV);
                return;

        case KEY_DC:    post_request (REQUEST_DELETE_AT);
                return;

        case KEY_ENTER:    post_request (REQUEST_DIALOG_ACCEPT);
                return;

        default:    if (! is_fkey (ch)) post_request_i (REQUEST_INSERT, ch);
                return;
        }
    } /* DIALOG_STRING */

    if (iscntrl (ch) && ! is_fkey (ch))
    {
        switch (my_unctrl (ch))
        {
        case 'L':   post_request (REQUEST_FORCE_REFRESH);
                    return;

        case 'X':   post_request (REQUEST_GRACEFUL_EXIT);
                    return;

        case 'S':   post_request (REQUEST_SAVE);
                    return;

        case 'D':   post_request (REQUEST_DUPLICATE_CARD);
                    return;

        case 'N':   post_request (REQUEST_NEW_CARD);
                    return;

        case 'F':   post_request (REQUEST_NEXT_CARD);
                    return;

        case 'B':   post_request (REQUEST_PREVIOUS_CARD);
                    return;

        case 'J':   post_request_i (REQUEST_INSERT, '\n');
                    return;

        case 'I':   post_request_i (REQUEST_INSERT, '\t');
                    return;

        case 'K':   post_request (REQUEST_DELETE_CARD);
                    return;

        case 'P':   post_request (REQUEST_PDF);
                    return;

        /* Due to legacy terminal issues with KEY_BACKSPACE */
        case '?':   post_request (REQUEST_DELETE_PREV);
                    return;

        default:    return;
        }
    }

    switch (ch)
    {
    case KEY_LEFT:    post_request_i (REQUEST_MOVE_CURSOR, LEFT);
            return;

    case KEY_UP:    post_request_i (REQUEST_MOVE_CURSOR, UP);
            return;

    case KEY_RIGHT:    post_request_i (REQUEST_MOVE_CURSOR, RIGHT);
            return;

    case KEY_DOWN:    post_request_i (REQUEST_MOVE_CURSOR, DOWN);
            return;

    case KEY_NPAGE:    post_request (REQUEST_NEXT_CARD);
            return;

    case KEY_PPAGE:    post_request (REQUEST_PREVIOUS_CARD);
            return;

    case KEY_BACKSPACE:    post_request (REQUEST_DELETE_PREV);
            return;

    case KEY_DC:    post_request (REQUEST_DELETE_AT);
            return;

    default:    if (! is_fkey (ch)) post_request_i (REQUEST_INSERT, ch);
            return;
    }
}
