#include "card.h"
#include "pdf.h"

#include<assert.h>
#include<stdio.h>
#include<stdlib.h>
#include<string.h>

enum ObjectType { NORMAL, PAGES, PAGE, RESOURCES };

struct PdfObject {
    struct PdfObject *next;
    int number;
    enum ObjectType type;
    char *final, *intermediate;
};

struct PdfFile {
    struct PdfObject *objects;
    char *data;
    long storage_size, storage_used;
};

struct PdfStream {
    char *data;
    int storage_size, storage_used;
};

/**************************** STATIC DECLARATIONS ****************************/
static void add_header (struct PdfStream *stream, char *filename, float x, float y);
static void add_card (struct PdfStream *stream, struct card *c, float x, float y);
static void add_footer (struct PdfStream *stream, int page, float x, float y);

static struct PdfFile *pdf_init (char *media_box);
static int pdf_write_file (char *filename, struct PdfFile *pdf);
static void pdf_free (struct PdfFile *pdf);
static void pdf_free_object (struct PdfObject *obj);
static int pdf_highest_object_number (struct PdfObject *obj);
static void pdf_add_objects_to_file_data (struct PdfFile *pdf, struct PdfObject *obj, long *offsets);

static void pdf_add_to_stream (struct PdfStream *stream, char *string);

static void pdf_resources (struct PdfFile *pdf, char *resource_string);
static void pdf_define_font (struct PdfFile *pdf, char *font_string);

static struct PdfStream *pdf_new_page (void);
static void pdf_Td (struct PdfStream *stream, float x_offset, float y_offset);
static void pdf_Tj (struct PdfStream *stream, char *text, int length);
static void pdf_Text (struct PdfStream *stream, char *text, int length);
static char *escape_control_chars (char *text, int length);
static void pdf_re (struct PdfStream *stream, float x, float y, float width, float height);
static void pdf_end_page (struct PdfFile *pdf, struct PdfStream *stream);
static void center_and_pad (char *buffer, char *src, int width);
/**************************** STATIC DECLARATIONS ****************************/

int pdf_write_one (struct card *c, char *crd_filename, char *pdf_filename, int letter_size)
{
    struct PdfFile *pdf;
    struct PdfStream *stream;

    pdf = pdf_init (letter_size ? "[0 0 612 792]" : "[0 0 595 842]");
    pdf_resources (pdf, "<</Font <</F1 %d 0 R>>>>");
    pdf_define_font (pdf, "<</Type /Font /Subtype /Type1 /BaseFont /Courier /Encoding /WinAnsiEncoding>>");

    stream = pdf_new_page ();
    pdf_add_to_stream (stream, "/F1 10.08 Tf\n");
    pdf_add_to_stream (stream, "0.24 w\n");

    if (letter_size)
    {
        add_header (stream, crd_filename, 69.0f, 750.16f);
        add_card (stream, c, 60.12f, 716.32f);
        add_footer (stream, 1, 69.0f, 46.24f);
    }
    else
    {
        add_header (stream, crd_filename, 67.0f, 790.16f);
        add_card (stream, c, 58.12f, 756.32f);
        add_footer (stream, 1, 67.0f, 40.16f);
    }

    pdf_end_page (pdf, stream);
    pdf_write_file (pdf_filename, pdf);
    pdf_free (pdf);

    return 1;
}

int pdf_write_all (struct card *c, char *crd_filename, char *pdf_filename, int letter_size)
{
    int page, card, total_cards;
    struct PdfFile *pdf;
    struct PdfStream *stream;

    pdf = pdf_init (letter_size ? "[0 0 612 792]" : "[0 0 595 842]");
    pdf_resources (pdf, "<</Font <</F1 %d 0 R>>>>");
    pdf_define_font (pdf, "<</Type /Font /Subtype /Type1 /BaseFont /Courier /Encoding /WinAnsiEncoding>>");

    for (total_cards = 0, page = 1; c; page++)
    {
        stream = pdf_new_page ();
        pdf_add_to_stream (stream, "/F1 10.08 Tf\n");
        pdf_add_to_stream (stream, "0.24 w\n");

        if (letter_size)
            add_header (stream, crd_filename, 69.0f, 750.16f);
        else
            add_header (stream, crd_filename, 67.0f, 790.16f);

        for (card = 0; card < (letter_size ? 4 : 5) && c; card++, total_cards++, c = c->next)
            if (letter_size)
                add_card (stream, c, 60.12f, 716.32f - card * 132.24f);
            else
                add_card (stream, c, 58.12f, 756.32f - card * 132.24f);

        if (letter_size)
            add_footer (stream, page, 69.0f, 46.24f);
        else
            add_footer (stream, page, 67.0f, 40.16f);

        pdf_end_page (pdf, stream);
    }

    pdf_write_file (pdf_filename, pdf);
    pdf_free (pdf);

    return total_cards;
}

static void add_header (struct PdfStream *stream, char *filename, float x, float y)
{
    char pad80[80];

    pdf_add_to_stream (stream, "BT\n");
    pdf_Td (stream, x, y);
    center_and_pad (pad80, filename, 80);
    pdf_Tj (stream, pad80, 80);
    pdf_add_to_stream (stream, "ET\n");
}


static void add_card (struct PdfStream *stream, struct card *c, float x, float y)
{
    char row_text[BODY_WIDTH];
    int row, row_len;

    pdf_re (stream, x, y, 255.12f, -126.24f);
    pdf_re (stream, x, y - 14.4f, 255.12f, 0.0f);
    pdf_add_to_stream (stream, "s\n");

    pdf_add_to_stream (stream, "BT\n");
    pdf_Td (stream, x + 9.12f, y - 12.48f);
    pdf_Tj (stream, c->title->string, (int) c->title->length);
    pdf_Td (stream, 0, -10.8f);
    get_row (c, row_text, &row_len, 0);
    pdf_Tj (stream, row_text, row_len);
    pdf_add_to_stream (stream, "10.08 TL\n");
    for (row = 1; ; row++)
    {
        if (! get_row (c, row_text, &row_len, row))
            break;
        pdf_Text (stream, row_text, row_len);
    }
    pdf_add_to_stream (stream, "ET\n");
}

static void add_footer (struct PdfStream *stream, int page, float x, float y)
{
    char buffer[50], pad80[80];

    pdf_add_to_stream (stream, "BT\n");
    pdf_Td (stream, x, y);
    sprintf (buffer, "Page %d", page);
    center_and_pad (pad80, buffer, 80);
    pdf_Tj (stream, pad80, 80);
    pdf_add_to_stream (stream, "ET\n");
}

/* call with "[0 0 595 842]" for A4 and "[0 0 612 792]" for letter */
static struct PdfFile *pdf_init (char *media_box)
{
    struct PdfFile *pdf;
    struct PdfObject *obj;

    pdf = malloc (sizeof (struct PdfFile));
    if (!pdf) exit (1);

    pdf->objects = NULL;
    pdf->data = strdup ("%PDF-1.0\n");
    pdf->storage_size = (int) strlen (pdf->data) + 1;
    pdf->storage_used = pdf->storage_size;

    obj = malloc (sizeof (struct PdfObject));
    if (!obj) exit (1);

    obj->next = NULL;
    pdf->objects = obj;

    obj->number = 1;
    obj->type = NORMAL;
    obj->final = strdup ("<</Type /Catalog /Pages 2 0 R>>");
    obj->intermediate = NULL;

    obj = malloc (sizeof (struct PdfObject));
    if (!obj) exit (1);

    obj->next = pdf->objects;
    pdf->objects = obj;

    obj->number = 2;
    obj->type = PAGES;
    obj->final = NULL;
    obj->intermediate = strdup (media_box);

    return pdf;
}

static int pdf_write_file (char *filename, struct PdfFile *pdf)
{
    FILE *f;
    int highest, i;
    long *offsets;

    if (!(f = fopen (filename, "wb")))
        return 0;

    highest = pdf_highest_object_number (pdf->objects);
    offsets = malloc ((size_t) (1 + highest) * sizeof (long));
    if (!offsets) exit (1);

    pdf_add_objects_to_file_data (pdf, pdf->objects, offsets);
    /* Don't write \0 */
    fwrite (pdf->data, (size_t) pdf->storage_used - 1, 1, f);

    fprintf (f, "xref\n0 %d\n0000000000 65535 f\n", highest);
    for (i = 1; i <= highest; i++)
        fprintf (f, "%010ld 00000 n\n", offsets[i]);
    fprintf (f, "trailer\n<< /Size %d /Root 1 0 R>>\nstartxref\n%ld\n%%%%EOF\n", highest + 1, pdf->storage_used - 1);
    fclose (f);

    return 1;
}

static void pdf_free (struct PdfFile *pdf)
{
    pdf_free_object (pdf->objects);
    free (pdf->data);
}

static void pdf_free_object (struct PdfObject *obj)
{
    /* The objects are stored as a reverse linked list */
    if (obj->next) pdf_free_object (obj->next);
    free (obj->final);
    free (obj->intermediate);
}

static int pdf_highest_object_number (struct PdfObject *obj)
{
    int a, b;

    if (! obj->next) return obj->number;

    a = obj->number;
    b = pdf_highest_object_number (obj->next);
    return a > b ? a : b;
}

static void pdf_add_objects_to_file_data (struct PdfFile *pdf, struct PdfObject *obj, long *offsets)
{
    int chars_printed;
    long bytes_to_add;
    char buffer[30];

    if (obj->next) pdf_add_objects_to_file_data (pdf, obj->next, offsets);

    switch (obj->type)
    {
        case NORMAL:
        case RESOURCES:
            assert (obj->intermediate == NULL);

            break;

        case PAGES:
        {
            int i, page_count, *page_objects;
            struct PdfObject *p;
            char buffer[50];

            assert (obj->final == NULL);

            for (page_count = 0, p = pdf->objects; p != NULL; p = p->next)
                if (p->type == PAGE) page_count++;

            page_objects = malloc ((size_t) page_count * sizeof (int));
            if (!page_objects) exit (1);

            for (i = 0, p = pdf->objects; p != NULL; p = p->next)
                if (p->type == PAGE)
                    page_objects[page_count - 1 - i++] = p->number;

            obj->final = malloc (100 + 20 * (size_t) page_count + strlen (obj->intermediate));
            if (!obj->final) exit (1);

            strcpy (obj->final, "<</Type /Pages /Kids [");
            for (i = 0; i < page_count; i++)
            {
                sprintf (buffer, "%d 0 R ", page_objects[i]);
                strcat (obj->final, buffer);
            }

            sprintf (buffer, "] /Count %d /MediaBox ", page_count);
            strcat (obj->final, buffer);
            strcat (obj->final, obj->intermediate);
            strcat (obj->final, ">>");

            free(obj->intermediate);
            obj->intermediate = NULL;

            break;
        }

        case PAGE:
        {
            int pages_obj, resources_obj;
            struct PdfObject *p;
            char buffer[200];

            assert (obj->final == NULL);

            for (pages_obj = -1, resources_obj = -1, p = pdf->objects; p != NULL; p = p->next)
                if (p->type == RESOURCES) resources_obj = p->number;
                else if (p->type == PAGES) pages_obj = p->number;

            assert (pages_obj != -1 && resources_obj != -1);

            if ((chars_printed = sprintf (buffer, "<</Type /Page /Parent %d 0 R /Resources %d 0 R /Contents %s>>", pages_obj, resources_obj, obj->intermediate)) < 0)
                exit (1);
            obj->final = malloc ((size_t) chars_printed + 1);
            strcpy (obj->final, buffer);
            if (!obj->final) exit (1);

            free (obj->intermediate);
            obj->intermediate = NULL;

            break;
        }
    }

    /* Account for '\0' which is not written */
    offsets[obj->number] = pdf->storage_used - 1;
    if ((chars_printed = sprintf (buffer, "%d 0 obj\n", obj->number)) < 0)
        exit (1);
    bytes_to_add = chars_printed + 8;
    bytes_to_add += strlen (obj->final);
    if (pdf->storage_size - pdf->storage_used < bytes_to_add)
    {
        pdf->storage_size = pdf->storage_size * 2 + bytes_to_add;
        pdf->data = realloc (pdf->data, (size_t) pdf->storage_size);
    }

    strcat (pdf->data, buffer);
    strcat (pdf->data, obj->final);
    strcat (pdf->data, "\nendobj\n");
    pdf->storage_used += bytes_to_add;
}

static void pdf_add_to_stream (struct PdfStream *stream, char *string)
{
    int bytes_to_add;

    bytes_to_add = (int) strlen (string);
    if (stream->storage_size - stream->storage_used < bytes_to_add)
    {
        stream->storage_size = stream->storage_size * 2 + bytes_to_add;
        stream->data = realloc (stream->data, (size_t) stream->storage_size);
    }

    strcat (stream->data, string);
    stream->storage_used += bytes_to_add;
}

/* resource_string should contain a %d for the font object number */
/* pdf_define_font should be called immediately after pdf_resources */
static void pdf_resources (struct PdfFile *pdf, char *resource_string)
{
    struct PdfObject *obj;

    obj = malloc (sizeof (struct PdfObject));
    if (!obj) exit (1);

    obj->next = pdf->objects;
    pdf->objects = obj;

    obj->number = obj->next->number + 1;
    obj->type = RESOURCES;
    obj->final = malloc (strlen (resource_string) + 30);
    sprintf (obj->final, resource_string, obj->number + 1);
    obj->intermediate = NULL;
}

static void pdf_define_font (struct PdfFile *pdf, char *font_string)
{
    struct PdfObject *obj;

    obj = malloc (sizeof (struct PdfObject));
    if (!obj) exit (1);

    obj->next = pdf->objects;
    pdf->objects = obj;

    obj->number = obj->next->number + 1;
    obj->type = NORMAL;
    obj->final = strdup (font_string);
    obj->intermediate = NULL;
}

static struct PdfStream *pdf_new_page (void)
{
    struct PdfStream *stream;

    stream = malloc (sizeof (struct PdfObject));
    if (!stream) exit (1);

    stream->data = malloc (200);
    if (!stream->data) exit (1);

    stream->data[0] = '\0';
    stream->storage_used = 1;
    stream->storage_size = 200;

    return stream;
}

static void pdf_Td (struct PdfStream *stream, float x_offset, float y_offset)
{
    char *buffer;

    buffer = malloc (100);
    if (!buffer) exit (1);

    sprintf (buffer, "%.2f %.2f Td\n", x_offset, y_offset);

    pdf_add_to_stream (stream, buffer);
    free (buffer);
}

static void pdf_Tj (struct PdfStream *stream, char *text, int length)
{
    char *buffer;

    buffer = escape_control_chars (text, length);
    strcat (buffer, "Tj\n");

    pdf_add_to_stream (stream, buffer);
    free (buffer);
}

static void pdf_Text (struct PdfStream *stream, char *text, int length)
{
    char *buffer;

    buffer = escape_control_chars (text, length);
    strcat (buffer, "'\n");

    pdf_add_to_stream (stream, buffer);
    free (buffer);
}

/* Returned buffer must be freed, includes surrounding parentheses and has at
   least 8 extra chars of space in buffer. */
static char *escape_control_chars (char *text, int length)
{
    char *buffer, *p, *q, octal[5];

    /* In case every character needs to be escaped */
    buffer = malloc ((unsigned) length * 4 + 10);
    if (!buffer) exit (1);

    buffer[0] = '(';
    for (p = buffer + 1, q = text; q < text + length; )
        if (*q >= 32 && *q <= 126 && *q != '(' && *q != ')' && *q != '\\')
            *p++ = *q++;
        else
        {
            sprintf (octal, "\\%03o", (unsigned char)*q++);
            memcpy (p, octal, 4);
            p += 4;
        }
    *p++ = ')';
    *p = '\0';
    return buffer;
}

static void pdf_re (struct PdfStream *stream, float x, float y, float width, float height)
{
    char *buffer;

    buffer = malloc (100);
    if (!buffer) exit (1);

    sprintf (buffer, "%.2f %.2f %.2f %.2f re\n", x, y, width, height);

    pdf_add_to_stream (stream, buffer);
    free (buffer);
}

static void pdf_end_page (struct PdfFile *pdf, struct PdfStream *stream)
{
    struct PdfObject *obj;
    char buffer[50];

    obj = malloc (sizeof (struct PdfObject));
    if (!obj) exit (1);

    obj->next = pdf->objects;
    pdf->objects = obj;

    obj->number = obj->next->number + 1;
    obj->type = PAGE;
    sprintf (buffer, "%d 0 R", obj->number + 1);
    obj->intermediate = strdup (buffer);
    obj->final = NULL;

    obj = malloc (sizeof (struct PdfObject));
    if (!obj) exit (1);

    obj->next = pdf->objects;
    pdf->objects = obj;

    obj->number = obj->next->number + 1;
    obj->type = NORMAL;
    obj->intermediate = NULL;
    obj->final = malloc ((size_t) stream->storage_used + 50);

    sprintf (obj->final, "<</Length %ld>>\nstream\n", (long) strlen (stream->data));
    strcat (obj->final, stream->data);
    strcat (obj->final, "endstream");

    free(stream->data);
    free(stream);
}

static void center_and_pad (char *buffer, char *src, int width)
{
    size_t center, left, right;

    center = strlen (src);

    if (center > (size_t) width)
    {
        memcpy (buffer, src, center);
        buffer[width] = '\0';
        return;
    }

    left = ((size_t) width - center) / 2;
    right = (size_t) width - left - center;

    memset (buffer, ' ', left);
    memcpy (buffer + left, src, center);
    memset (buffer + left + center, ' ', right);
}

