#include "card.h"
#include "cardfile.h"
#include "fileio.h"
#include "pdf.h"
#include "state.h"
#include "requests.h"

#include <assert.h>
#include <curses.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static int min (int a, int b);
static int max (int a, int b);
static void put_at (int x, int y, chtype ch);
static int print_horizontal_line (int x1, int x2, int y, chtype ch);
static int print_vertical_line (int x, int y1, int y2, chtype ch);
static void print_at_length (int x, int y, char *s, int len);
static void print_at (int x, int y, char *s);
static void state_refresh (struct state *s);
static void state_save_accept (struct state *s);
static void state_string_return_to_menu (struct state *s);
static void state_graceful_exit_yes (struct state *s);
static void state_return_to_menu (struct state *s);
static void state_pdf_all_cards (struct state *s);
static void state_pdf_one_card (struct state *s);
static void state_pdf_letter (struct state *s);
static void state_pdf_a4 (struct state *s);
static void state_pdf_accept (struct state *s);

static int min (int a, int b)
{
    return a < b ? a : b;
}

static int max (int a, int b)
{
    return a > b ? a : b;
}

struct state *state_new (void)
{
    struct state *s = malloc (sizeof (struct state));
    if (!s) exit(1);

    s->cur_card = NULL;
    s->filename = NULL;
    s->screen_rows = s->screen_cols = 0;
    s->first_visible = s->last_visible = NULL;
    s->status = MENU;
    s->input_mode = TITLE_MODE;

    s->status_text = NULL;
    s->dialog_prompt = NULL;
    s->dialog_cursor_offset = 0;
    s->dialog_yes = NULL;
    s->dialog_no = NULL;
    s->dialog_abort = NULL;
    s->dialog_accept = NULL;
    s->dialog_input = NULL;

    return s;
}

/* Assumes first, last are both valid or NULL */
void state_determine_visible_titles (struct state *s)
{
    int shown_titles = s->screen_rows - 1;

    /* If called early before curses has been set up */
    if (s->screen_rows == 0)
        return;

    if (s->first_visible == NULL)
    {
        s->first_visible = s->cur_card;
        s-> last_visible = cards_forward (s->cur_card, shown_titles - 1);
    }
    else
    {
        /* Scroll screen if needed */
        if (is_after (s->cur_card, s->first_visible))
        {
            s->first_visible = s->cur_card;
            s-> last_visible = cards_forward (s->cur_card, shown_titles - 1);
        }

        if (is_after (s->last_visible, s->cur_card))
        {
            s->first_visible = cards_backward (s->cur_card, shown_titles - 1);
            s-> last_visible = s->cur_card;
        }

        /* If any cards were deleted, grow on-screen window to fit */
        /* Grow down, and then grow up if needed */
        if (count_cards_later (s->last_visible, s->first_visible) < shown_titles - 1)
        {
            s-> last_visible =  cards_forward (s->first_visible, shown_titles - 1);
            s->first_visible = cards_backward (s->last_visible,  shown_titles - 1);
        }

        /* If the screen shrunk, reduce on-screen window to fit */
        /* Trim from the bottom, then from the top if needed */
        if (count_cards_later (s->last_visible, s->first_visible) > shown_titles - 1)
        {
            int top_rows = count_cards_later (s->cur_card,  s->first_visible);

            if (top_rows + 1 < shown_titles)
                /* We can trim only from the bottom */
                s-> last_visible = cards_forward (s->first_visible, shown_titles - 1);
            else
            {
                /* We must trim all the bottom and some from the top */
                s-> last_visible = s->cur_card;
                s->first_visible = cards_backward (s->last_visible,  shown_titles - 1);
            }
        }
    }
}

void state_duplicate_card (struct state *s)
{
    assert (s->cur_card && s->first_visible && s->last_visible);
    duplicate_card (s->cur_card);
    state_jump_card (s, s->cur_card->prev, REFRESH_OK);
    state_determine_visible_titles (s);
    state_refresh (s);
}

void state_read_cardfile (struct state *s, char *filename)
{
    FILE *f;

    assert (! s->filename);
    assert (! s->cur_card);

    if (s->status_text)
        free (s->status_text);
    s->status_text = malloc(200);
    if (!s->status_text) exit (1);
    s->status = TEMP_STATUS;

    if (!filename)
    {
        s->cur_card = alloc_blank_card ();
        s->filename = editstring_new_from_cstr ("", AUTORESIZE);
        sprintf (s->status_text, "New file");
        return;
    }

    s->filename = editstring_new_from_cstr (filename, AUTORESIZE);

    if ((f = fopen (s->filename->string, "rb")))
    {
        s->cur_card = read_cardfile (f);
        fclose (f);

        if (! s->cur_card)
        {
            fputs ("error parsing cardfile\n", stderr);
            exit(1);
        }

        sort_cardfile (s->cur_card);
        s->cur_card = head (s->cur_card);

        if (any_objects_skipped (s->cur_card))
            sprintf (s->status_text, "Loaded %d cards - WARNING: IMAGES OR OLE"
                "OBJECTS DISCARDED", count_cards (s->cur_card));
        else
            sprintf (s->status_text, "Loaded %d cards", count_cards (s->cur_card));
    }
    else
    {
        s->cur_card = alloc_blank_card ();
        sprintf (s->status_text, "Error opening cardfile");
    }

    state_determine_visible_titles (s);
}

void state_insert (struct state *s, char ch)
{
    assert (s->cur_card && s->first_visible && s->last_visible);

    switch (s->input_mode)
    {
        case BODY_MODE:
        {
            insert_body_char (s->cur_card, ch);
            state_status_update (s);
            break;
        }
        case TITLE_MODE:
        {
            if (ch == '\n')
            {
                s->input_mode = BODY_MODE;
                reset_cursor_pos (s->cur_card, SET_BODY_TO_TITLE);

                state_status_update (s);
                break;
            }

            insert_title_char (s->cur_card, ch);

            sort_cardfile (s->cur_card);
            state_determine_visible_titles (s);

            state_status_update (s);
            break;
        }
        case DIALOG_YN:
            /* Should not happen */
            break;
        case DIALOG_STRING:
            editstring_insert_char (s->dialog_input, ch);
            s->dialog_cursor_offset = min (s->screen_cols - (int) strlen (s->dialog_prompt), s->dialog_cursor_offset + 1);
            break;
    }

    state_refresh (s);
}

void state_previous_card (struct state *s, int refresh_ok)
{
    assert (s->cur_card && s->first_visible && s->last_visible);

    state_status_update (s);

    if (s->cur_card->prev)
        state_jump_card (s, s->cur_card->prev, refresh_ok);
    else
       if (refresh_ok)
            state_refresh (s);
}

void state_next_card (struct state *s, int refresh_ok)
{
    assert (s->cur_card && s->first_visible && s->last_visible);

    state_status_update (s);

    if (s->cur_card->next)
        state_jump_card (s, s->cur_card->next, refresh_ok);
    else
       if (refresh_ok)
            state_refresh (s);
}

void state_jump_card (struct state *s, struct card *c, int refresh_ok)
{
    s->cur_card = c;
    reset_cursor_pos (s->cur_card, SET_BOTH_TO_ZERO);
    state_determine_visible_titles (s);
    if (refresh_ok)
        state_refresh (s);
}

void state_move_cursor (struct state *s, int dir)
{
    assert (s->cur_card && s->first_visible && s->last_visible);

    if (s->input_mode == TITLE_MODE)
    {
        if (dir == LEFT)
            editstring_advance_cursor_pos (s->cur_card->title, -1);
        if (dir == RIGHT)
            editstring_advance_cursor_pos (s->cur_card->title,  1);
        if (dir == DOWN) {
            s->input_mode = BODY_MODE;
            reset_cursor_pos (s->cur_card, SET_BODY_TO_TITLE);
        }

        state_status_update (s);
        state_refresh (s);
        return;
    }

    if (s->input_mode == DIALOG_YN)
    {
        /* Should not happen */
        return;
    }

    if (s->input_mode == DIALOG_STRING)
    {
        if (dir == LEFT)
        {
            if (editstring_advance_cursor_pos (s->dialog_input, -1))
                s->dialog_cursor_offset = max (0, s->dialog_cursor_offset - 1);
            state_refresh (s);
        }
        if (dir == RIGHT)
        {
            if (editstring_advance_cursor_pos (s->dialog_input, 1))
                s->dialog_cursor_offset = min (s->screen_cols - (int) strlen (s->dialog_prompt), s->dialog_cursor_offset + 1);
            state_refresh (s);
        }
        return;
    }

    /* s->input_mode == BODY_MODE */
    if (dir == LEFT)
        multiedit_cursor_left (s->cur_card->body);
    if (dir == RIGHT)
        multiedit_cursor_right (s->cur_card->body);
    if (dir == UP)
    {
        if (multiedit_apparent_cursor_row (s->cur_card->body) == 0)
        {
            s->input_mode = TITLE_MODE;
            reset_cursor_pos (s->cur_card, SET_TITLE_TO_BODY);
        }
        else
            multiedit_cursor_up (s->cur_card->body);
    }
    if (dir == DOWN)
        multiedit_cursor_down (s->cur_card->body);

    state_status_update (s);
    state_refresh (s);
}

void state_new_card (struct state *s)
{
    struct card *new_card;

    assert (s->cur_card && s->first_visible && s->last_visible);

    new_card = prepend_blank_card (s->cur_card);
    if (! new_card)
        exit (1);
    s->input_mode = TITLE_MODE;

    state_status_update (s);
    state_jump_card (s, new_card, REFRESH_OK);
}

void state_delete_card (struct state *s)
{
    struct card *new_card;

    assert (s->cur_card && s->first_visible && s->last_visible);

    /* don't delete only card */
    if (s->first_visible == s->last_visible)
        return;

    /* this is the card to jump to; prev if it exists, else next */
    new_card = delete_card (s->cur_card);
    assert(new_card);

    /* don't allow first_visible or last_visible to dangle */
    if (s->cur_card == s->first_visible)
        s->first_visible = new_card;
    else if (s->cur_card == s->last_visible)
        s->last_visible = new_card;

    state_status_update (s);
    state_jump_card (s, new_card, REFRESH_OK);
}

static void put_at (int x, int y, chtype ch)
{
    int my, mx, by, bx;

    getmaxyx (stdscr, my, mx);
    getbegyx (stdscr, by, bx);
    if (x < bx || x >= mx || y < by || y >= my)
        return;

    move (y, x);
    addch (ch);
}

static int print_horizontal_line (int x1, int x2, int y, chtype ch)
{
    int a, b;

    for (a = min (x1, x2), b = max (x1, x2); a <= b; a++)
        put_at (a, y, ch);

    return 1;
}

static int print_vertical_line (int x, int y1, int y2, chtype ch)
{
    int a, b;

    for (a = min (y1, y2), b = max (y1, y2); a <= b; a++)
        put_at (x, a, ch);

    return 1;
}

static void print_at_length (int x, int y, char *s, int len)
{
    for (; len; len--)
        put_at (x + len-1, y, (unsigned char) s[len-1]);
}

static void print_at (int x, int y, char *s)
{
    print_at_length (x, y, s, (int) strlen (s));
}

void state_force_refresh (struct state *st)
{
    state_determine_visible_titles (st);
    clear ();
    clearok (curscr, 1);
    state_refresh (st);
    clearok (curscr, 0);
}

static void state_refresh (struct state *st)
{
    int a, b, x;
    int p, r, t, y;
    int height;

    char text_buf[BODY_WIDTH + 1];
    int i, length;

    struct card *card;

    assert (st->cur_card && st->first_visible && st->last_visible);

    bkgdset (BKGD_ATTR | ' ');
    erase ();

    /* Drawing the title, card, and lines

       (a,p)--------(b,p)
         |            |
       (a,r)--------(b,r)
         |            |
         |            |
       (a,t)--------(b,t) */


    /* Calculate the corner coordinates */

    height = 1 + 1 + 1 + BODY_HEIGHT + 1;

    a = 0;
    b = 1 + max (TITLE_WIDTH, BODY_WIDTH)-1 + 1;
    p = (st->screen_rows - height) / 2;
    r = p + 1 + 1;
    t = r + BODY_HEIGHT + 1;

    /* Draw the lines */

    attrset (CARD_ATTR);
    print_horizontal_line (a, b, p, ACS_HLINE);
    print_horizontal_line (a, b, r, ACS_HLINE);
    print_horizontal_line (a, b, t, ACS_HLINE);
    print_vertical_line (a, p, r, ACS_VLINE);
    print_vertical_line (b, p, r, ACS_VLINE);
    print_vertical_line (a, r, t, ACS_VLINE);
    print_vertical_line (b, r, t, ACS_VLINE);
    put_at (a, p, ACS_ULCORNER);
    put_at (b, p, ACS_URCORNER);
    put_at (a, r, ACS_LTEE);
    put_at (b, r, ACS_RTEE);
    put_at (a, t, ACS_LLCORNER);
    put_at (b, t, ACS_LRCORNER);

    /* Draw the title

       +-------------+
       |(x,y)        |
       +-------------+
       |             |
       |             |
       +-------------+ */

    x = a + 1;
    y = p + 1;

    attrset (TEXT_ATTR);
    print_horizontal_line (x, x+TITLE_WIDTH-1, y, ' ');
    print_at_length (x, y, st->cur_card->title->string, (int) st->cur_card->title->length);

    /* Draw the text

       +-------------+
       |             |
       +-------------+
       |(x,y)        |
       |             |
       +-------------+ */

    x = a + 1;
    y = r + 1;

    for (i = 0; i < BODY_HEIGHT; i++)
    {
        memset (text_buf, ' ', BODY_WIDTH);
        get_row (st->cur_card, text_buf, &length, i);
        print_at_length (x, y + i, text_buf, BODY_WIDTH);
    }

    /* Draw the titles

            (x,y)
       +------+
       +------+
       +------+

                       */

    x = b + 2;
    y = 0;

    card = st->first_visible;
    do
    {
        if (card == st->cur_card)
            attrset (HILITE_ATTR);
        else
            attrset (INDEX_ATTR);

        print_horizontal_line (x, st->screen_cols-1, y, ' ');
        print_at_length (x, y, card->title->string, (int) card->title->length);

        card = card->next;
        y++;
    }
    while (card != st->last_visible->next);

    /* Status */

    attrset (HILITE_ATTR);
    print_horizontal_line (0, st->screen_cols-1, st->screen_rows-1, ' ');

    switch (st->status)
    {
    case MENU:
        print_at (0, st->screen_rows-1,
        "Ctrl- eXit Save Dupe New Forward Back Kill Pdf");
        break;
    case TEMP_STATUS:
        print_at (0, st->screen_rows-1, st->status_text);
        break;
    case DIALOG:
        /* In DIALOG_INPUT mode we also output part of dialog_input, see below */
        print_at (0, st->screen_rows-1, st->dialog_prompt);
        break;
    }

    /* Position cursor */

    switch (st->input_mode)
    {
        case BODY_MODE:
        {
            x = a + 1;
            y = r + 1;

            x += multiedit_apparent_cursor_col (st->cur_card->body);
            y += multiedit_apparent_cursor_row (st->cur_card->body);
            move (y, x);
            break;
        }
        case TITLE_MODE:
        {
            x = a + 1;
            y = p + 1;

            x += st->cur_card->title->cursor_pos;
            move (y, x);
            break;
        }
        case DIALOG_YN:
        {
            x = (int) strlen (st->dialog_prompt);
            y = st->screen_rows-1;

            move (y, x);
            break;
        }
        case DIALOG_STRING:
        {
            char *ss;
            int prompt_len = (int) strlen (st->dialog_prompt);

            st->dialog_cursor_offset = min (st->screen_cols - prompt_len - 1, st->dialog_cursor_offset);
            ss = editstring_relative_substr (st->dialog_input, st->dialog_cursor_offset, st->screen_cols - prompt_len);
            print_at (prompt_len, st->screen_rows - 1, ss);

            free (ss);

            x = prompt_len + st->dialog_cursor_offset;
            y = st->screen_rows-1;

            move (y, x);
            break;
        }
    }

    refresh ();
}

static void state_save_accept (struct state *s)
{
    FILE *f;

    assert (s->cur_card);

    editstring_free (s->filename);
    s->filename = s->dialog_input;

    if (s->status_text)
        free (s->status_text);
    s->status_text = malloc(200);
    if (!s->status_text) exit (1);
    s->status = TEMP_STATUS;

    f = fopen (s->filename->string, "wb");
    if (!f)
    {
        sprintf (s->status_text, "Error opening file for writing");
    }
    else
    {
        if (write_cardfile (f, s->cur_card))
            sprintf (s->status_text, "Saved %d cards", count_cards (s->cur_card));
        else
            sprintf (s->status_text, "Error saving cards");
        fclose (f);
    }

    s->input_mode = s->old_mode;

    s->dialog_prompt = NULL;
    s->dialog_input = NULL;
    s->dialog_accept = NULL;
    s->dialog_abort = NULL;

    state_refresh (s);
}

static void state_string_return_to_menu (struct state *s)
{
    s->dialog_prompt = NULL;
    editstring_free (s->dialog_input);
    s->dialog_input = NULL;
    s->dialog_accept = NULL;
    s->dialog_abort = NULL;
    s->input_mode = s->old_mode;
    s->status = MENU;

    state_refresh (s);
}

void state_save (struct state *s)
{
    s->dialog_prompt = "File name to write: ";
    s->dialog_accept = state_save_accept;
    s->dialog_abort = state_string_return_to_menu;
    s->old_mode = s->input_mode;
    s->dialog_input = editstring_copy (s->filename);
    s->dialog_input->cursor_pos = (int) s->dialog_input->length;
    /* Will be clamped if necessary if the filename is long by state_refresh */
    s->dialog_cursor_offset = (int) s->dialog_input->length;
    s->input_mode = DIALOG_STRING;
    /* Should flush any requests in the queue except maybe resize/refresh */
    s->status = DIALOG;

    state_refresh (s);
}

void state_delete_prev (struct state *s)
{
    switch (s->input_mode)
    {
        case BODY_MODE:
            delete_body_prev (s->cur_card);
            state_status_update (s);
            break;
        case TITLE_MODE:
            delete_title_prev (s->cur_card);
            sort_cardfile (s->cur_card);
            state_determine_visible_titles (s);

            state_status_update (s);
            break;
       case DIALOG_YN:
            /* Should not happen */
            break;
       case DIALOG_STRING:
            if (s->dialog_input->cursor_pos <= s->dialog_cursor_offset)
                s->dialog_cursor_offset = max (0, s->dialog_cursor_offset - 1);
            editstring_delete_prev (s->dialog_input);
            break;
    }
    state_refresh (s);
}

void state_delete_at (struct state *s)
{
    switch (s->input_mode)
    {
        case BODY_MODE:
            delete_body_at (s->cur_card);
            state_status_update (s);
            break;
        case TITLE_MODE:
            delete_title_at (s->cur_card);
            sort_cardfile (s->cur_card);
            state_determine_visible_titles (s);

            state_status_update (s);
            break;
       case DIALOG_YN:
            /* Should not happen */
            break;
       case DIALOG_STRING:
            editstring_delete_at (s->dialog_input);
            /* No change to dialog_cursor_offset */
            break;
    }
    state_refresh (s);
}

void state_free (struct state *s)
{
    assert (s->cur_card);
    delete_cardfile (s->cur_card);
    if (s->filename) editstring_free(s->filename);
    /* Assume s->dialog_prompt is not dynamically allocated */
    if (s->dialog_input) editstring_free(s->dialog_input);
    free (s);
}

void state_inform_width (struct state *s, int w)
{
    s->screen_cols = w;
}

void state_inform_height (struct state *s, int h)
{
    s->screen_rows = h;
}

void state_dialog_yes (struct state *s)
{
    s->dialog_yes(s);
}

void state_dialog_no (struct state *s)
{
    s->dialog_no(s);
}

void state_dialog_abort (struct state *s)
{
    s->dialog_abort(s);
}

void state_dialog_accept (struct state *s)
{
    s->dialog_accept(s);
}

static void state_graceful_exit_yes (struct state *s)
{
    s->dialog_prompt = NULL;
    s->dialog_yes = NULL;
    s->dialog_no = NULL;
    s->dialog_abort = NULL;
    s->input_mode = s->old_mode;
    s->status = MENU;

    post_request (REQUEST_EXIT_NOW);
}

static void state_return_to_menu (struct state *s)
{
    s->dialog_prompt = NULL;
    s->dialog_yes = NULL;
    s->dialog_no = NULL;
    s->dialog_abort = NULL;
    s->input_mode = s->old_mode;
    s->status = MENU;

    state_refresh (s);
}

void state_graceful_exit (struct state *s)
{
    if (any_cards_dirty(s->cur_card))
    {
        s->dialog_prompt = "Exit without saving? ";
        s->dialog_yes = state_graceful_exit_yes;
        s->dialog_no = state_return_to_menu;
        s->dialog_abort = state_return_to_menu;
        s->old_mode = s->input_mode;
        s->input_mode = DIALOG_YN;
        /* Should flush any requests in the queue except maybe resize/refresh */
        s->status = DIALOG;

        state_refresh (s);
    }
    else
        post_request (REQUEST_EXIT_NOW);
}

/* Small state machine for the status line at the bottom of the screen */
void state_status_update (struct state *s)
{
    assert(s);
    assert(s->status != DIALOG);

    if (s->status == TEMP_STATUS)
    {
        if (s->status_text)
        {
            free (s->status_text);
            s->status_text = NULL;
        }

        s->status = MENU;
    }
}

void state_pdf (struct state *s)
{
    s->dialog_prompt = "Pdf entire cardfile (y), instead of current card only (n)? ";
    s->dialog_yes = state_pdf_all_cards;
    s->dialog_no = state_pdf_one_card;
    s->dialog_abort = state_return_to_menu;
    s->old_mode = s->input_mode;
    s->input_mode = DIALOG_YN;
    /* Should flush any requests in the queue except maybe resize/refresh */
    s->status = DIALOG;

    state_refresh (s);
}

static void state_pdf_all_cards (struct state *s)
{
    s->pdf_state = PDF_ALL;

    s->dialog_prompt = "Pdf letter paper (y), instead of A4 (n)? ";
    s->dialog_yes = state_pdf_letter;
    s->dialog_no = state_pdf_a4;
    s->dialog_abort = state_return_to_menu;

    state_refresh (s);
}

static void state_pdf_one_card (struct state *s)
{
    s->pdf_state = PDF_ONE;

    s->dialog_prompt = "Pdf letter paper (y), instead of A4 (n)? ";
    s->dialog_yes = state_pdf_letter;
    s->dialog_no = state_pdf_a4;
    s->dialog_abort = state_return_to_menu;

    state_refresh (s);
}

static void state_pdf_letter (struct state *s)
{
    s->pdf_state |= PDF_LETTER;

    s->dialog_yes = NULL;
    s->dialog_no = NULL;

    s->dialog_prompt = "File name to write: ";
    s->dialog_accept = state_pdf_accept;
    s->dialog_abort = state_string_return_to_menu;
    s->dialog_input = editstring_new_from_cstr ("", 1);
    s->dialog_input->cursor_pos = (int) s->dialog_input->length;
    s->dialog_cursor_offset = (int) s->dialog_input->length;
    s->input_mode = DIALOG_STRING;
    s->status = DIALOG;

    state_refresh (s);
}

static void state_pdf_a4 (struct state *s)
{
    s->pdf_state |= PDF_A4;

    s->dialog_yes = NULL;
    s->dialog_no = NULL;

    s->dialog_prompt = "File name to write: ";
    s->dialog_accept = state_pdf_accept;
    s->dialog_abort = state_string_return_to_menu;
    s->dialog_input = editstring_new_from_cstr ("", 1);
    s->dialog_input->cursor_pos = (int) s->dialog_input->length;
    s->dialog_cursor_offset = (int) s->dialog_input->length;
    s->input_mode = DIALOG_STRING;
    s->status = DIALOG;

    state_refresh (s);
}

static void state_pdf_accept (struct state *s)
{
    int ret;

    assert (s->cur_card);

    if (s->status_text)
        free (s->status_text);
    s->status_text = malloc(200);
    if (!s->status_text) exit (1);
    s->status = TEMP_STATUS;

    ret = s->pdf_state & PDF_ALL ? pdf_write_all (head (s->cur_card), s->filename->string, s->dialog_input->string, s->pdf_state & PDF_LETTER) : pdf_write_one (s->cur_card, s->filename->string, s->dialog_input->string, s->pdf_state & PDF_LETTER);

    if (ret)
    {
        sprintf (s->status_text, "Successfully made pdf of %d card%s", ret, ret > 1 ? "s" : "");
    }
    else
    {
        sprintf (s->status_text, "Error making pdf file");
    }

    editstring_free (s->dialog_input);
    s->dialog_input = NULL;

    s->input_mode = s->old_mode;

    s->dialog_prompt = NULL;
    s->dialog_accept = NULL;
    s->dialog_abort = NULL;

    state_refresh (s);
}
