#include <assert.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#include "editstr.h"

struct EditString *editstring_new (int max_string_length, int autoresizes)
{
    struct EditString *ms = malloc (sizeof (struct EditString));
    if (!ms) exit(1);

    ms->storage_size = max_string_length + 1;
    ms->string = malloc ((size_t) ms->storage_size);
    if (!ms->string) exit(1);

    memset (ms->string, 0, (size_t) ms->storage_size);

    ms->length = 0;
    ms->autoresizes = autoresizes;

    ms->cursor_pos = 0;

    return ms;
}

struct EditString *editstring_copy (struct EditString *old)
{
    struct EditString *ms = malloc (sizeof (struct EditString));
    if (!ms) exit(1);

    *ms = *old;

    ms->string = malloc ((size_t) ms->storage_size);
    if (!ms->string) exit (1);

    memcpy (ms->string, old->string, (size_t) ms->storage_size);

    return ms;
}

struct EditString *editstring_new_from_cstr (char *cstr, int autoresizes)
{
    int len = (int) strlen (cstr);
    struct EditString *ms = editstring_new (len, autoresizes);
    strcpy (ms->string, cstr);
    ms->length = len;
    return ms;
}

void editstring_free (struct EditString *ms)
{
    if (!ms) return;
    free (ms->string);
    free (ms);
}

char *editstring_relative_substr (struct EditString *ms, int back, int len)
{
    int start = ms->cursor_pos < back ? 0 : ms->cursor_pos - back;
    char *ret = calloc ((size_t) len + 1, 1);
    if (!ret) exit (1);

    strncpy (ret, ms->string + start, (size_t) len);
    return ret;
}

int editstring_insert_char (struct EditString *ms, char ch)
{
    char *tail_head;
    int tail_length;

    if (ms->length == (ms->storage_size - 1))
    {
        if (!ms->autoresizes)
            return 0;
        else
        {
            int newsize = ms->storage_size * 2;
            char *newstr = calloc ((size_t) newsize, 1);
            if (!newstr) exit (1);

            memcpy (newstr, ms->string, (size_t) ms->length + 1);
            free (ms->string);
            ms->string = newstr;
            ms->storage_size = newsize;
        }
    }

    tail_head = ms->string + ms->cursor_pos;
    tail_length = ms->length - ms->cursor_pos;

    memmove (tail_head + 1, tail_head, (size_t) tail_length);
    ms->string[ms->cursor_pos] = ch;
    ms->length++;
    ms->cursor_pos++;

    return 1;
}

/* Deletes ms->string[ms->cursor_pos-1] and shifts the tail left one slot */
int editstring_delete_prev (struct EditString *ms)
{
    char *tail_head = ms->string + ms->cursor_pos;
    int tail_length = ms->length - ms->cursor_pos;

    if (ms->cursor_pos == 0) return 0;

    memmove (tail_head - 1, tail_head, (size_t) tail_length);
    ms->length--;
    ms->string[ms->length] = '\0';
    ms->cursor_pos--;

    return 1;
}

/* Deletes ms->string[ms->cursor_pos] and shifts the tail left one slot */
int editstring_delete_at (struct EditString *ms)
{
    char *tail_head = ms->string + ms->cursor_pos + 1;
    int tail_length = ms->length - ms->cursor_pos - 1;

    if (ms->cursor_pos == ms->length) return 0;

    memmove (tail_head - 1, tail_head, (size_t) tail_length);
    ms->length--;
    ms->string[ms->length] = '\0';
    ms->cursor_pos--;

    return 1;
}

int editstring_set_cstr (struct EditString *ms, char *cstr)
{
    return editstring_set_cstr_with_length (ms, cstr, (int) strlen (cstr));
}

/* len does not include nul terminator - in case source does not have one */
int editstring_set_cstr_with_length (struct EditString *ms, char *cstr, int len)
{
    if (len + 1 > ms->storage_size)
    {
        if (!ms->autoresizes)
            return 0;
        else
        {
            int newsize = ms->storage_size * 2;
            char *newstr = calloc ((size_t) newsize, 1);
            if (!newstr) exit (1);

            free (ms->string);
            ms->string = newstr;
            ms->storage_size = newsize;
        }
    }

    memcpy (ms->string, cstr, (size_t) len);
    ms->string[len] = '\0';
    ms->length = len;
    return 1;
}

/* Expands storage to hold string of length len, whether or not
   autoresizes is true */
int editstring_grow (struct EditString *ms, int len)
{
    if (len < ms->length)
        return 0;

    if (len + 1 > ms->storage_size)
    {
        char *newstr = calloc ((size_t) len + 1, 1);
        if (!newstr) exit (1);

        memcpy (newstr, ms->string, (size_t) ms->length + 1);
        free (ms->string);
        ms->string = newstr;
        ms->storage_size = len + 1;
    }

    return 1;
}

int editstring_advance_cursor_pos (struct EditString *ms, int change)
{
    /* Can point at one past the end of the string */
    if (ms->cursor_pos + change < 0 || ms->cursor_pos + change > ms->length) return 0;

    ms->cursor_pos += change;
    return 1;
}

int editstring_set_cursor_pos (struct EditString *ms, int pos) {
    /* Can point at one past the end of the string */
    if (pos < 0 || pos > ms->length) return 0;

    ms->cursor_pos = pos;
    return 1;
}
