#include "multied.h"

#include <assert.h>
#include <stdlib.h>
#include <string.h>

/* DOS-style carriage-return linefeed */
#define TEXT_NEWLINE "\x0d\x0a"
#define TEXT_NEWLINE_LENGTH 2

/**************************** STATIC DECLARATIONS ****************************/
static int column_difference (char *start_pos, char *q);
static char *pos_of_column (char *start_pos, int c);
static int row_from_location (struct MultiEdit *m, char *p);
static int col_from_location (struct MultiEdit *m, char *p);
static char *locate_position (struct MultiEdit *m, int col, int row);

static int num_rows_unlimited (struct MultiEdit *m);

static int insert_newline (struct MultiEdit *m);
static int insert_not_newline (struct MultiEdit *m, char ch);
static int delete_at (struct MultiEdit *m, char *p, int num_del);

static char *get_row_pos (struct MultiEdit *m, int row, char **end_pos);
static int get_next_row_pos (struct MultiEdit *m, char **pos, char *max_pos);
static int is_whitespace (char ch);
static int get_text_end_pos_next_row_pos (struct MultiEdit *m, char *buffer, char *max_pos, char **te_pos, char **nr_pos);

static void *my_memmem (void *haystack, int haystack_length, void *needle, int needle_length);

static int max_row (struct MultiEdit *m);

/**************************** STATIC DECLARATIONS ****************************/

static int column_difference (char *start_pos, char *q)
{
    int current_column;
    char *pos;
    
    for (current_column = 0, pos = start_pos; pos < q; pos++)
        current_column += *pos == '\t' ? SPACES_PER_TAB - current_column % SPACES_PER_TAB : 1;

    return current_column;
}

/* assumes c >= 0 */
static char *pos_of_column (char *start_pos, int c)
{
    int current_column;
    char *pos;

    for (current_column = 0, pos = start_pos; current_column < c; pos++)
        current_column += *pos == '\t' ? SPACES_PER_TAB - current_column % SPACES_PER_TAB : 1;

    if (c % 8 && *(pos-1) == '\t') pos--;
    return pos;
}

static int row_from_location (struct MultiEdit *m, char *p)
{
    int i;
    char *row_pos, *max;

    row_pos = m->text;
    max = m->text + m->text_length - 1;

    /* Allow cursor one past end of last row */
    assert (row_pos <= p && p <= max + 1);

    for (i = 0; i < m->control_height; i++)
    {
        if (get_next_row_pos (m, &row_pos, max))
        {
            if (p < row_pos)
                return i;
        }
        else
            /* Is in last row */
            return i;
    }

    assert (0);
}

static int col_from_location (struct MultiEdit *m, char *p)
{
    int i;
    char *row_pos, *max;

    row_pos = m->text;
    max = m->text + m->text_length - 1;

    /* Allow cursor one past end of last row */
    assert (row_pos <= p && p <= max + 1);

    for (i = 0; i < m->control_height; i++)
    {
        char *prev_row = row_pos;
        if (get_next_row_pos (m, &row_pos, max))
        {
            if (p < row_pos)
                return column_difference (prev_row, p);
        }
        else
            /* Is in last row */
            return column_difference (prev_row, p);
    }

    assert (0);
}

static char *locate_position (struct MultiEdit *m, int col, int row)
{
    int i, max_col;
    char *row_pos, *max, *start, *end;

    if (row < 0 || (col < 0 && row == 0))
        return NULL;

    row_pos = m->text;
    /* It is permissible to put the cursor one byte past the end of the text,
       provided this is within the text array */
    max = m->text + m->text_length - 1 + (m->text_length < m->max_text_bytes);

    for (i = 0; i < row; i++)
        if (! get_next_row_pos (m, &row_pos, max))
            /* Row doesn't exist */
            return NULL;

    start = row_pos;
    get_text_end_pos_next_row_pos (m, row_pos, max, &end, &row_pos);
    max_col = column_difference (start, end + 1) - 1;

    /* Allow position at end of row */
    if (max_col + 1 < col)
        /* Col doesn't exist */
        return NULL;

    return pos_of_column (start, col);
}

/* insert_body_not_newline temporarily may have one too many rows
   reduce error-checking */
static int num_rows_unlimited (struct MultiEdit *m)
{
    int i;
    char *pos, *max;

    pos = m->text;
    max = m->text + m->text_length - 1;

    for (i = 1; i < m->max_text_bytes / TEXT_NEWLINE_LENGTH; i++)
        if (! get_next_row_pos (m, &pos, max))
            /* Row doesn't exist */
            return i;

    /* Too many rows - corrupt card data. But some versions of
       Windows cardfile can produce such a file. */
    return m->control_height - 1;
}

/* Newlines take two characters, so we need a special function */
/* Assumes m->true_cursor_col has been normalized */
static int insert_newline (struct MultiEdit *m)
{
    char *start, *new_start;
    int tail_length;

    if (m->text_length + TEXT_NEWLINE_LENGTH > m->max_text_bytes)
        return 0;

    /* We don't know if it will fit until we try */

    start = locate_position (m, m->true_cursor_col, m->true_cursor_row);
    new_start = start + TEXT_NEWLINE_LENGTH;
    /* [0 .. start-1] [start .. text_length-1] */
    tail_length = (int) (m->text + m->text_length - start);

    memmove (new_start, start, (size_t) tail_length);
    memcpy (start, TEXT_NEWLINE, TEXT_NEWLINE_LENGTH);
    m->text_length += TEXT_NEWLINE_LENGTH;

    if (num_rows_unlimited (m) > m->control_height)
    {
        int i;

        /* It wrapped, take the character out */
        memmove (start, new_start, (size_t) tail_length);

        for (i = 0; i < TEXT_NEWLINE_LENGTH; i++)
            /* Mark deleted char with % for debugging */
            m->text[m->text_length-i-1] = '%';

        m->text_length -= TEXT_NEWLINE_LENGTH;
        return 0;
    }

    return 1;
}

/* Assumes m->true_cursor_col has been normalized */
static int insert_not_newline (struct MultiEdit *m, char ch)
{
    char *start;
    int tail_length;

    if (m->text_length + 1 > m->max_text_bytes)
        return 0;

    /* We don't know if it will fit until we try */

    start = locate_position (m, m->true_cursor_col, m->true_cursor_row);
    /* [0 .. start-1] [start .. text_length-1] */
    tail_length = (int) (m->text + m->text_length - start);

    memmove (start + 1, start, (size_t) tail_length);
    *start = ch;
    m->text_length++;

    if (num_rows_unlimited (m) > m->control_height)
    {
        /* It wrapped, take the character out */
        memmove (start, start + 1, (size_t) tail_length);

        /* Mark deleted char with % for debugging */
        m->text[m->text_length-1] = '%';

        m->text_length--;
        return 0;
    }

    return 1;
}

/* Returns success */
static int delete_at (struct MultiEdit *m, char *p, int num_del)
{
    char *max;
    int tail_length;

    max = m->text + m->text_length - 1;
    tail_length = (int) (max - p - num_del + 1);

    /* [0 .. p-1] [p .. p+num_del-1] [p+num_del .. text_length-1] */
    memmove (p, p + num_del, (size_t) tail_length);

    /* Set deleted character to % for debugging */
    memset (max-num_del + 1, '%', (size_t) num_del);
    m->text_length -= num_del;

    return 1;
}

static char *get_row_pos (struct MultiEdit *m, int row, char **end_pos) {
    char *row_pos = m->text, *next_row_pos;
    char *max_pos = m->text + m->text_length - 1;
    int i = 0;

    while (1)
    {
        get_text_end_pos_next_row_pos (m, row_pos, max_pos, end_pos, &next_row_pos);
        if (i++ == row)
            return row_pos;
        row_pos = next_row_pos;
    }
}

/* Returns success */
static int get_next_row_pos (struct MultiEdit *m, char **pos, char *max_pos)
{
    char *text_end;

    return get_text_end_pos_next_row_pos (m, *pos, max_pos, &text_end, pos);
}

static int is_whitespace (char ch)
{
    return memchr ("\0\t\n\r ", ch, 5) != NULL;
}

/* Returns success */
static int get_text_end_pos_next_row_pos (struct MultiEdit *m, char *buffer, char *max_pos, char **te_pos, char **nr_pos)
{
    int search_length, row_length, i;
    char *ww_pos = NULL, *ww_next_pos = NULL;
    
    search_length = (int) (max_pos - buffer + 1);
    
    for (*te_pos = NULL, i = 0, row_length = 0; i < search_length && row_length < m->control_width; i++)
    {
        /* Track the first character after a word break */
        if (ww_pos && is_whitespace (buffer[i-1]))
            ww_next_pos = buffer + i;

        if (is_whitespace (buffer[i]))
            /* In a word break, so remember the last character before the word break */
            ww_pos = *te_pos;
        else
            /* Not in a word break */
            *te_pos = buffer + i;

        /* row_length < m->control_width, therefore not a hidden tab */
        if (buffer[i] == '\t')
        {
            /* Assumes m->control_width is divisible by SPACES_PER_TAB */
            row_length += SPACES_PER_TAB - (row_length % SPACES_PER_TAB);
            continue;
        }
        
        /* Does this line ends because it has a newline? */
        if (search_length - i >= TEXT_NEWLINE_LENGTH && memcmp (buffer + i, TEXT_NEWLINE, TEXT_NEWLINE_LENGTH) == 0)
        {
            if (search_length - i >= TEXT_NEWLINE_LENGTH)
            {
                /* Update te_pos to include any non-hidden whitespace on the line */
                *te_pos = buffer + i - 1;
                if (nr_pos) *nr_pos = buffer + i + TEXT_NEWLINE_LENGTH;

                return 1;
            }
            else
            {
                /* Update te_pos to include any non-hidden whitespace on the line */
                *te_pos = buffer + i - 1;

                /* There's no room for another row after this one */
                return 0;
            }
        }

        /* buffer[i] is a normal character */
        row_length++;
    }

    /* Did not find a newline. Either we counted off control_width columns
       (row_length == m->control_width) or we ran out of data
       (i == search_length). Or both. */
    
    assert (row_length <= m->control_width);

    if (ww_pos != NULL && !is_whitespace (buffer[i - 1]) && (i - 1 < max_pos - buffer) && !is_whitespace (buffer[i]))
    {
       /* We saw a word break (ww_pos is not NULL), the last character in the
       row is not whitespace, we're not at the end of the buffer and the next
       character is not whitespace either (if it is whitespace, it is hidden
       whitespace). So word-wrap. */
       
       if (nr_pos) *nr_pos = ww_next_pos;
       *te_pos = ww_pos;
       return 1;
    }

    /* *te_pos points to the last text character in the row (or is NULL if
       none). Update to include any non-hidden whitespace on the line. But
       still return NULL if the line is empty (don't let te_pos point
       outside the buffer) */
    *te_pos = i == 0 ? NULL : buffer + i - 1;

    /* Since this is a full line, skip over hidden tabs and/or hidden newline */
    if (row_length == m->control_width)
    {
        for (; i < search_length; i++)
            if (buffer[i] == '\t')
                continue;
            else if (search_length - i >= TEXT_NEWLINE_LENGTH && memcmp (buffer + i, TEXT_NEWLINE, TEXT_NEWLINE_LENGTH) == 0)
            {
                i += TEXT_NEWLINE_LENGTH;
                break;
            } else
                break;
    }
    
    if (i - 1 == max_pos - buffer)
        /* There's no room for another row after this one */
        return 0;
    
    if (nr_pos) *nr_pos = buffer + i;
    return 1;
}

/* Some systems have memmem, others don't */
static void *my_memmem (void *haystack, int haystack_length, void *needle, int needle_length)
{
    if (needle_length == 0)
    	return haystack;

    while (haystack_length >= needle_length)
    {
    	if (memcmp (haystack, needle, (size_t) needle_length) == 0)
    		return haystack;

    	haystack = ((char *) haystack) + 1;
    	haystack_length--;
    }

    return NULL;
}

static int max_row (struct MultiEdit *m)
{
    int i;
    char *row_pos, *max_pos;

    row_pos = m->text;
    max_pos = m->text + m->text_length - 1;

    for (i = 0; i < m->control_height; i++)
        if (! get_next_row_pos (m, &row_pos, max_pos))
            /* Row doesn't exist */
            return i;

    /* Too many rows - corrupt card data */
    assert (0);
}

struct MultiEdit *multiedit_new (int control_width, int control_height, int max_text_bytes)
{
    struct MultiEdit *m = malloc (sizeof (struct MultiEdit));
    if (!m) exit(1);

    m->max_text_bytes = max_text_bytes;
    m->text = malloc ((size_t) m->max_text_bytes);
    if (!m->text) exit(1);

    memset (m->text, 0, (size_t) m->max_text_bytes);

    m->text_length = 0;
    m->control_width = control_width;
    m->control_height = control_height;

    m->true_cursor_col = 0;
    m->true_cursor_row = 0;

    return m;
}

struct MultiEdit *multiedit_copy (struct MultiEdit *old)
{
    struct MultiEdit *m = malloc (sizeof (struct MultiEdit));
    if (!m) exit(1);

    *m = *old;

    m->text = malloc ((size_t) m->max_text_bytes);
    if (!m->text) exit(1);

    memcpy (m->text, old->text, (size_t) m->max_text_bytes);

    return m;
}

void multiedit_free (struct MultiEdit *m)
{
    if (!m) return;
    free (m->text);
    free (m);
}

/* Returns row_len measured in bytes, excluding newline and/or hidden whitespace */
int multiedit_get_row_raw (struct MultiEdit *m, char *buffer, int *row_len, int row)
{
    char *row_pos, *end;

    if (row > max_row (m))
        return 0;

    row_pos = get_row_pos (m, row, &end);
    *row_len = (int)(end ? end - row_pos + 1 : 0);

    memcpy (buffer, row_pos, (size_t) *row_len);
    return 1;
}

int multiedit_get_row (struct MultiEdit *m, char *buffer, int *row_len, int row)
{
    int i;

    if (! multiedit_get_row_raw (m, buffer, row_len, row))
        return 0;

    for (i = 0; i < *row_len;)
    {
        int spaces_to_add, room_left;
        char *tab_pos, *tail_start, *new_start;
        int tail_length;

        if (buffer[i] != '\t')
        {
            i++;
            continue;

        }

        /* Assumes m->control_width is divisible by SPACES_PER_TAB.
           spaces_to_add will always be at least 1 */
        spaces_to_add = SPACES_PER_TAB - (i % SPACES_PER_TAB);
        room_left = m->control_width - *row_len + 1;

        assert (room_left >= spaces_to_add);

        tab_pos = buffer + i;
        tail_start = tab_pos + 1;
        new_start = tab_pos + spaces_to_add;
        tail_length = *row_len - i - 1;

        memmove (new_start, tail_start, (size_t) tail_length);
        memset (tab_pos, ' ', (size_t) spaces_to_add);

        i += spaces_to_add;
        *row_len += spaces_to_add - 1;
    }
    
    return 1;
}

int multiedit_apparent_cursor_col (struct MultiEdit *m)
{
    char *row_pos, *end, *pos;
    int max_col, clipped_col;
    
    row_pos = get_row_pos (m, m->true_cursor_row, &end);
    max_col = end ? column_difference (row_pos, end + 1) : 0;

    /* Ensure that column is at most one position past end of row */
    clipped_col = max_col < m->true_cursor_col ? max_col : m->true_cursor_col;

    /* If column is within a tab, return position of next character */
    pos = pos_of_column (row_pos, clipped_col);
    return column_difference (row_pos, pos);
}

int multiedit_apparent_cursor_row (struct MultiEdit *m)
{
    return m->true_cursor_row;
}

int multiedit_insert_char (struct MultiEdit *m, char ch)
{
    char *old_pos;
    
    /* Normalize cursor */
    m->true_cursor_col = multiedit_apparent_cursor_col (m);
    old_pos = locate_position (m, m->true_cursor_col, m->true_cursor_row);

    if (ch == '\n' ? insert_newline (m) : insert_not_newline (m, ch))
    {
        char *new_pos = old_pos + (ch == '\n' ? TEXT_NEWLINE_LENGTH : 1);
        m->true_cursor_col = col_from_location (m, new_pos);
        m->true_cursor_row = row_from_location (m, new_pos);

        return 1;
    }
    return 0;
}

int multiedit_delete_prev (struct MultiEdit *m)
{
    char *p;
    int num_del;

    /* Normalize cursor */
    m->true_cursor_col = multiedit_apparent_cursor_col (m);
    p = locate_position (m, m->true_cursor_col, m->true_cursor_row);

    assert (p);

    /* Are we deleting a newline, or a single character? */
    num_del = p - TEXT_NEWLINE_LENGTH >= m->text && memcmp (p - TEXT_NEWLINE_LENGTH, TEXT_NEWLINE, TEXT_NEWLINE_LENGTH) == 0 ? TEXT_NEWLINE_LENGTH : 1;

    if (p - num_del >= m->text && delete_at (m, p - num_del, num_del))
    {
        m->true_cursor_row = row_from_location (m, p - num_del);
        m->true_cursor_col = col_from_location (m, p - num_del);
        return 1;
    }
    else
        return 0;
}

int multiedit_delete_at (struct MultiEdit *m)
{
    char *p, *max;
    int num_del;

    /* Normalize cursor */
    m->true_cursor_col = multiedit_apparent_cursor_col (m);
    p = locate_position (m, m->true_cursor_col, m->true_cursor_row);

    assert (p);

    /* Are we deleting a newline, or a single character? */
    max = m->text + m->text_length - 1;
    num_del = p == my_memmem (p, (int) (max - p + 1), TEXT_NEWLINE, TEXT_NEWLINE_LENGTH)
        ? TEXT_NEWLINE_LENGTH : 1;

    return delete_at (m, p, num_del);
}

void multiedit_cursor_left (struct MultiEdit *m)
{
    char *p;
    int num_skip;

    /* Normalize cursor */
    m->true_cursor_col = multiedit_apparent_cursor_col (m);
    p = locate_position (m, m->true_cursor_col, m->true_cursor_row);

    /* Are we skipping over a newline, or a normal character? */
    num_skip = p - TEXT_NEWLINE_LENGTH >= m->text && memcmp (p - TEXT_NEWLINE_LENGTH, TEXT_NEWLINE, TEXT_NEWLINE_LENGTH) == 0 ? TEXT_NEWLINE_LENGTH : 1;

    if (p > m->text)
    {
        m->true_cursor_row = row_from_location (m, p - num_skip);
        m->true_cursor_col = col_from_location (m, p - num_skip);
    }
}

void multiedit_cursor_right (struct MultiEdit *m)
{
    char *p, *last_pos, *buffer_end;
    int num_skip;

    /* Normalize cursor */
    m->true_cursor_col = multiedit_apparent_cursor_col (m);
    p = locate_position (m, m->true_cursor_col, m->true_cursor_row);
    last_pos = m->text + m->text_length - 1;
    buffer_end = m->text + m->max_text_bytes - 1;

    /* Are we skipping over a newline, or a normal character? */
    num_skip = last_pos - p + 1 >= TEXT_NEWLINE_LENGTH && memcmp (p, TEXT_NEWLINE, TEXT_NEWLINE_LENGTH) == 0 ? TEXT_NEWLINE_LENGTH : 1;

    if (p < buffer_end && p <= last_pos)
    {
        m->true_cursor_row = row_from_location (m, p + num_skip);
        m->true_cursor_col = col_from_location (m, p + num_skip);
    }
}

void multiedit_cursor_up (struct MultiEdit *m)
{
    if (m->true_cursor_row > 0)
        m->true_cursor_row--;
}

void multiedit_cursor_down (struct MultiEdit *m)
{
    if (m->true_cursor_row < max_row (m))
        m->true_cursor_row++;
}

int multiedit_set_cursor_row (struct MultiEdit *m, int row) {
    if (row < 0 || row > max_row (m)) return 0;

    m->true_cursor_row = row;
    return 1;
}

int multiedit_set_cursor_col (struct MultiEdit *m, int col) {
    /* Can be set beyond the end of a row - will be normalized later */
    if (col < 0 || col > m->control_width) return 0;

    m->true_cursor_col = col;
    return 1;
}
